//======================================================================
// 
// GuardFileDriver.c
//
// Copyright (C) CVIC Software Engineering CO.,LTD.
//
//======================================================================
#include "Define.h"
#include "GuardFileDriver.h"
#include "../include/Driver/Ioctlcmd.h"
#include "FastIo.h"
#include "Irp.h"
#include "Disk.h"
#include "Policy.h"
#include "Filter.h"
#include "Event.h"


//----------------------------------------------------------------------
//                           DEFINES
//----------------------------------------------------------------------
//
// This is GuardFileDriver's user-inteface device object. It is addressed
// by calls from the ManagerApp including CreateFile and DeviceIoControl
//
PDEVICE_OBJECT      gUIDevice;

//
// This is our Driver Object
//
PDRIVER_OBJECT      gGuardFileDriverDriver;


//
// This is the offset into a KPEB of the current process name. This is determined
// dynamically by scanning the process block belonging to the GUI for the name
// of the system process, in who's context we execute in DriverEntry
//
ULONG               gProcessNameOffset;


// the current version of Windows NT/Windows 2000
ULONG	MajorVersion = 0;
ULONG	MinorVersion = 0;



VOID InitGuardFileDriver()
{
    InitIrp();
	InitPolicy();
	InitEvent();
	InitFilter();
}

VOID UnInitGuardFileDriver()
{
	//
    // Now we can free any memory that is allocatedp
    //
	ClearFilter();
	ClearPolicy();
    HashCleanup();
	ClearEvent();
}

//----------------------------------------------------------------------
//
// GuardFileDriverUnload
//
// Our job is done - time to leave.
//
//----------------------------------------------------------------------
VOID GuardFileDriverUnload( IN PDRIVER_OBJECT DriverObject )
{
	UNICODE_STRING         deviceLinkUnicodeString;

	DbgPrint(("GuardFileDriver: GuardFileDriverUnload-start\n"));
	//
	// Disconnect from file systems
	//
	HookDriveSet( 0, DriverObject );

	//
	// Delete the symbolic link for our GUI device
	//
	RtlInitUnicodeString( &deviceLinkUnicodeString, Device_Link_Name_String );
	IoDeleteSymbolicLink( &deviceLinkUnicodeString );

	//
	// Delete the device object, making sure that the GUI device
	// object is always deleted.
	//
	if ( gUIDevice == DriverObject->DeviceObject )
	{
		IoDeleteDevice( DriverObject->DeviceObject );
	}
	else
	{
		IoDeleteDevice( DriverObject->DeviceObject );
		IoDeleteDevice( gUIDevice ); 
	}

	UnInitGuardFileDriver();

	DbgPrint(("GuardFileDriver: GuardFileDriverUnload-end\n"));
}


//----------------------------------------------------------------------
//
// DriverEntry
//
// Installable driver initialization. Here we just set ourselves up.
//
//----------------------------------------------------------------------
NTSTATUS DriverEntry(IN PDRIVER_OBJECT DriverObject, IN PUNICODE_STRING RegistryPath )
{
	NTSTATUS                ntStatus;
	UNICODE_STRING          deviceNameUnicodeString;
	UNICODE_STRING          deviceLinkUnicodeString;
	ULONG                   i;

	DbgPrint (("GuardFileDriver: DriverEntry-start\n"));

	gGuardFileDriverDriver = DriverObject;

	// Get OS version
	PsGetVersion(&MajorVersion, &MinorVersion, NULL, NULL);
	//DbgPrint(("GuardFileDriver: OS MajorVer[%ld] MinorVer[%ld]\n", MajorVersion, MinorVersion));

	//
	// If not NT 4.0 Final Release, shorten the Fast I/O table so that GuardFileDriver 
	// will work on the Betas and Release Candidates
	//
	if( *NtBuildNumber < NT4FINAL )
	{
		FastIOHook.SizeOfFastIoDispatch = (ULONG)((MYULONG) &FastIOHook.FastIoQueryNetworkOpenInfo - 
			(MYULONG) &FastIOHook);
	} 

	//    
	// Setup the device name
	//    
	RtlInitUnicodeString (&deviceNameUnicodeString, Device_Name_String );

	//
	// Create the device used for ManagerApp communications
	//
	ntStatus = IoCreateDevice ( DriverObject,
								0,
								&deviceNameUnicodeString,
								FILE_DEVICE_FILEDRIVER,
								0,
								TRUE,
								&gUIDevice );

	//
	// If successful, make a symbolic link that allows for the device
	// object's access from Win32 programs
	//
	if (NT_SUCCESS(ntStatus))
	{
		//
		// Create a symbolic link that the UI can specify to gain access
		// to this driver/device
		//
		RtlInitUnicodeString (&deviceLinkUnicodeString, Device_Link_Name_String );
		ntStatus = IoCreateSymbolicLink (&deviceLinkUnicodeString, &deviceNameUnicodeString );
		if (!NT_SUCCESS(ntStatus))
		{
			DbgPrint (("GuardFileDriver: ----IoCreateSymbolicLink failed\n"));
		}

		for( i = 0; i <= IRP_MJ_MAXIMUM_FUNCTION; i++ )
		{
			DriverObject->MajorFunction[i] = GuardFileDriverDispatch;
		}

  		DriverObject->DriverUnload                          = GuardFileDriverUnload;

		//
		// Set up the Fast I/O dispatch table
		//
		DriverObject->FastIoDispatch = &FastIOHook;
	}

	//
	// If something went wrong, cleanup the device object and don't load
	//
	if (!NT_SUCCESS(ntStatus))
	{
		DbgPrint(("GuardFileDriver: ----Failed to create our device!\n"));

		if( gUIDevice )
		{
			IoDeleteDevice( gUIDevice );
		}
		return ntStatus;
	}

	InitGuardFileDriver();

	DbgPrint(("GuardFileDriver: DriverEntry-end(ok)\n"));
	return ntStatus;
}
